import logging
from typing import Any, Callable, Coroutine, Dict, List, Optional, Union

from services import types
from sqlalchemy import MetaData, create_engine, event, inspect
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.ext.asyncio import (AsyncSession, async_scoped_session,
                                    create_async_engine)
from sqlalchemy.ext.asyncio.engine import AsyncEngine
# from sqlalchemy.dialects.postgresql.base import PGInspector
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from .utils import drop_everything


class SQL:

    def __init__(self, db: types.Database,
                 set_session_factory=True,
                 autoflush=True,
                 init_engine=True,
                 future=True):
        """ SQL sync helper to build connections and sessions
        it's prepared to be used with Core and ORM

        :param db: accepts type :class:`services.types.Database`
        :type db: services.types.Database
        """
        self.conf = db
        self._future = future
        self._autoflush = autoflush
        if init_engine:
            self._init(session_factory=set_session_factory)

    @classmethod
    def from_uri(cls, uri: str, init_engine=True) -> "SQL":
        _db = types.Database(sync_url=uri)
        return cls(_db, init_engine=init_engine)

    def _init(self, session_factory=True):
        if "sqlite" in self.conf.sync_url.split("://", maxsplit=1)[0]:
            self._engine = create_engine(self.conf.sync_url,
                                         echo=self.conf.debug)
        else:
            self._engine = create_engine(
                self.conf.sync_url,
                pool_size=self.conf.pool_size,
                max_overflow=self.conf.max_overflow,
                echo=self.conf.debug
            )

        if session_factory:
            self._factory = self.sessionmaker(
                autoflush=self._autoflush, future=self._future)

    @property
    def engine(self) -> Engine:
        return self._engine

    def create_all(self, meta: MetaData):
        """
        To create a database it's need the MetaData object
        from sqlalchemy
        """
        meta.create_all(self.engine)

    def drop_all(self, meta: Optional[MetaData] = None, all_=True):
        """ it drops the content of a database, if all_ is set False
        then it only will drop elementes in meta.

        """

        if all_:
            drop_everything(self.engine)
        elif meta:
            meta.drop_all(self.engine)
        else:
            raise AttributeError(
                "MetaData object or all_ param should be provided")

    def list_tables(self) -> List[str]:
        inspector = inspect(self.engine)
        return inspector.get_table_names()

    def reflect_metadata(self) -> MetaData:
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        return metadata

    def add_event(self, func: Callable, type_event="connect"):
        event.listen(self.engine, type_event, func)

    def sessionmaker(self, autoflush=True, future=True) -> sessionmaker:
        """ wrapper around SqlAlchemy sessionmaker

        :param future: default True then Query 2.0 style
        """
        return sessionmaker(bind=self.engine,
                            autoflush=autoflush,
                            future=future)

    def session_factory(self) -> Session:
        """ it's actually returns a session """
        return self._factory()

    def scoped_session(self, autoflush=True, future=True) -> Session:
        """ wrapper around SqlAlchemy session_scoped

        Create a single scoped_session registry when the web application
        first starts, ensuring that this object is accessible by the
        rest of the application.

        ensures that scoped_session.remove() is called when the web
        request ends.

        :param future: default True then Query 2.0 style
        """
        return scoped_session(sessionmaker(bind=self.engine,
                                           autoflush=autoflush,
                                           future=future))


class AsyncSQL:
    """
    As example checks:
    https://docs.sqlalchemy.org/en/14/_modules/examples/asyncio/async_orm.html

    scoped_session, useful in thread environments is removed in Async,
    AsyncSession is prefered insted: https://docs.sqlalchemy.org/en/14/orm/extensions/asyncio.html#using-asyncio-scoped-session
    """

    # Meta = MetaData()

    def __init__(self, db: types.Database,
                 set_session_factory=True,
                 autoflush=True,
                 future=True):
        self.conf = db
        self._future = future
        self._autoflush = autoflush
        self._set_session_factory = set_session_factory

    async def init(self):
        if "sqlite" in self.conf.async_url.split("://", maxsplit=1)[0]:
            self._engine = create_async_engine(self.conf.async_url,
                                               echo=self.conf.debug)
        else:
            self._engine = create_async_engine(
                self.conf.async_url,
                pool_size=self.conf.pool_size,
                max_overflow=self.conf.max_overflow,
                echo=self.conf.debug
            )

        if self._set_session_factory:
            self._factory = self.sessionmaker(
                autoflush=self._autoflush, future=self._future)

    @classmethod
    def from_uri(cls, uri: str, init_engine=False) -> "AsyncSQL":
        _db = types.Database(async_url=uri)
        return cls(_db)

    def sessionmaker(self,
                     autoflush=True,
                     future=True,
                     expire_on_commit=False) -> sessionmaker:
        """
        expire_on_commit=False will prevent attributes from being expired
        after commit
        """
        return sessionmaker(
            self._engine,
            future=future,
            autoflush=autoflush,
            expire_on_commit=expire_on_commit, class_=AsyncSession
        )

    def session_factory(self) -> AsyncSession:
        """ it's actually returns a session """
        return self._factory()

    @property
    def engine(self) -> AsyncEngine:
        return self._engine

    @property
    def sync_engine(self) -> Engine:
        return self._engine.sync_engine

    async def create_all(self, meta: MetaData):
        async with self._engine.begin() as conn:
            await conn.run_sync(meta.create_all)

    async def drop_all(self, meta: Optional[MetaData] = None, all_=True):
        async with self._engine.begin() as conn:
            if all_:
                await conn.run_sync(
                    lambda sync_conn: drop_everything(sync_conn)
                )
            elif meta:
                await conn.run_sync(
                    lambda sync_conn: meta.drop_all(sync_conn)
                )
            else:
                raise AttributeError(
                    "MetaData object or all_ param should be provided")

    async def list_tables(self) -> Coroutine[Any, Any, List[str]]:
        async with self._engine.connect() as conn:
            tables = await conn.run_sync(
                lambda sync_conn: inspect(sync_conn).get_table_names()
            )
        return tables



from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm, Prompt

from services.jtemplates import render_to_file
from services.utils import get_parent_folder, mkdir_p, normalize_name

console = Console()
FOLDERS = [
    "server_conf",
]


def _empty_file(filename):
    with open(filename, "w", encoding="utf-8") as f:
        pass
    return True


def ask_webapp_name() -> str:
    parent = get_parent_folder()
    _default = normalize_name(parent)
    project_name = Prompt.ask(
        f"Write a name for default web app [yellow]please, avoid spaces and capital "
        "letters[/yellow]: ",
        default=_default,
    )
    name = normalize_name(project_name)
    console.print(
        f"The final name for the project is: [bold magenta]{name}[/bold magenta]"
    )
    return name


def final_words(project_name):
    p = Panel.fit(
        "[bold magenta]:smile_cat: Congrats!!!"
        f" Project [cyan]{project_name}[/cyan] created[/bold magenta]",
        border_style="red",
    )
    console.print(p)

    console.print(
        " [bold magenta]To test if everything is working "
        " you can run the following command:[/]\n"
    )
    console.print("\t[bold] srv web -L[/]\n")


def create_default(root, default_app):
    """
    Entrypoint when calling `srv startproject .`
    It will creates all the neccesary for a project to start
    """

    data = {"app_name": default_app}
    for f in FOLDERS:
        mkdir_p(f)

    _empty_file(f"{root}/server_conf/__init__.py")
    render_to_file(template="settings.py",
                   dst=f"{root}/server_conf/settings.py",
                   data=data)
    render_to_file(template="app/alembic.ini", dst=f"{root}/alembic.ini")
    create_app(root, default_app, init=True)

    final_words(default_app)


def alembic_files(root, app_name):
    # alembic specific
    dst = f"{root}/{app_name}"
    mkdir_p(f"{dst}/migrations/versions")
    data = {"app_name": app_name}

    render_to_file(template="alembic/env.py",
                   dst=f"{dst}/migrations/env.py",
                   data=data)
    render_to_file(template="alembic/script.py.mako",
                   dst=f"{dst}/migrations/script.py.mako")
    with open(f"{root}/alembic.ini", "a", encoding="utf-8") as f:
        f.write((
            "\n\n"
            f"[{app_name}]\n"
            "sqlalchemy.url = \n"
            f"script_location = %(here)s/{app_name}/migrations/\n"
            f"models_module = {app_name}.models\n"
            f"version_table = { app_name }_version\n"
            f"db_name = default\n"
        ))


def create_app(root, app_name, init=False):
    dst = f"{root}/{app_name}"
    mkdir_p(dst)

    data = {"app_name": app_name, "init": init}
    # creates and generates app's package files
    _empty_file(f"{dst}/__init__.py")
    render_to_file(template="app/web.py", dst=f"{dst}/web.py", data=data)
    render_to_file(template="app/views_bp.py",
                   dst=f"{dst}/views_bp.py", data=data)
    render_to_file(template="app/models.py", dst=f"{dst}/models.py", data=data)
    render_to_file(template="app/db.py",
                   dst=f"{dst}/db.py", data=data)

    if init:
        render_to_file(template="app/users_bp.py",
                       dst=f"{dst}/users_bp.py", data=data)

    alembic_files(root, app_name)


def create_app_old(root, app_name):
    dst = f"{root}/{app_name}"
    mkdir_p(dst)
    data = {"app_name": app_name}
    _empty_file(f"{dst}/__init__.py")
    render_to_file(template="app/web.py", dst=f"{dst}/web.py", data=data)
    render_to_file(template="app/views_bp.py",
                   dst=f"{dst}/views_bp.py", data=data)
    render_to_file(template="app/models.py", dst=f"{dst}/models.py", data=data)
    render_to_file(template="app/db.py",
                   dst=f"{dst}/db.py", data=data)

    mkdir_p(f"{dst}/migrations/versions")
    render_to_file(template="alembic/env.py",
                   dst=f"{dst}/migrations/env.py",
                   data=data)
    render_to_file(template="alembic/script.py.mako",
                   dst=f"{dst}/migrations/script.py.mako")
    with open(f"{root}/alembic.ini", "a", encoding="utf-8") as f:
        f.write((
            "\n\n"
            f"[{app_name}]\n"
            "sqlalchemy.url = \n"
            f"script_location = %(here)s/{app_name}/migrations/\n"
            f"models_module = {app_name}.models\n"
            f"version_table = { app_name }_version\n"
        ))



from sanic import Blueprint, Request
from sanic.response import json
from sanic_ext import openapi
from services.errors import AuthValidationFailed, WebAuthFailed
from services.security import protected
from services.types import JWTResponse, UserLogin
from services.users import get_users_mg

users_bp = Blueprint("users_api", url_prefix="users", version="v1")


@users_bp.post("/login")
@openapi.response(200, {"application/json": JWTResponse})
@openapi.response(403, dict(msg=str), "Not Found")
@openapi.body(UserLogin)
async def login_handler(request: Request):
    manager = get_users_mg(request)
    try:
        creds = UserLogin(**request.json)
        user = await manager.authenticate(username=creds.username,
                                          password=creds.password)
    except AuthValidationFailed as exc:
        raise WebAuthFailed() from exc

    jwt = await manager.generate_token(user)
    return json(jwt.dict(), 200)


@users_bp.get("/verify")
@openapi.response(200, {"application/json": JWTResponse})
@protected()
async def verify_handler(request: Request):
    return json(request.ctx.token_data, 200)


@users_bp.post("/refresh_token")
@openapi.response(200, {"application/json": JWTResponse})
@openapi.body(JWTResponse)
async def refresh_handler(request):
    if not request.app.config.AUTH_ALLOW_REFRESH:
        return json(dict(msg="Not found"), 404)

    at = request.token
    rftkn = request.json.get("refresh_token")
    if not rftkn:
        raise WebAuthFailed()

    old_token = JWTResponse(access_token=at, refresh_token=rftkn)
    manager = get_users_mg(request)
    # redis = request.ctx.web_redis
    try:
        jwt_res = await manager.refresh_token(old_token)
        return json(jwt_res.dict(), 200)
    except AuthValidationFailed as e:
        raise WebAuthFailed() from e



import asyncio
import importlib
import sys
from getpass import getpass
from pathlib import Path

import click
from rich.console import Console
from rich.prompt import Prompt
from services import conf, defaults, init_script, types
from services.db import SQL, AsyncSQL, Migration
from services.db.sqlhelper import AsyncSQL as AsyncSQL2
from services.users.managers import UserManager
from services.utils import get_class, mkdir_p

console = Console()


async def create_user(settings: types.Settings, data):
    db = AsyncSQL2(settings.DATABASES[settings.USER_DB])
    await db.init()
    # session = db.sessionmaker()
    manager = UserManager(db, model_class=settings.USER_MODEL,
                          salt=settings.SECURITY.AUTH_SALT)
    user = await manager.create(data)
    return user


@click.group(name="manager")
def managercli():
    """
    Managment tasks like user creation and db init
    """
    pass


@managercli.command(name="db")
# @click.option("--database", "-d", default="default", help="Database to apply migration")
@click.option("--message", "-m", default=None, help="Revision name")
@click.option("--rev-id", "-R", help="Revision id")
@click.option("--to", "-t", default="head", help="Revision to upgrade or downgrade")
@click.option("--head", default=None, help="Specify head revision or <branchname>@head to base new revision.")
@click.option("--branch-label", "-b", default=None, help="Specify a branch label to apply to the new revision")
@click.option("--name", "-n", default=None, help="package name where migration is located")
@click.option("--depends-on", "-d", help="list of 'depends on' identifiers")
@click.argument("action", type=click.Choice(["create", "drop", "upgrade", "downgrade", "revision"]))
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE, help="Fullpath to settings module")
def dbcli(action, message, rev_id, to, head, branch_label, name, depends_on, settings_module):
    """Create or Drop tables from a database"""
    settings = conf.load_conf(settings_module)
    db = SQL(settings.SQL)
    if action == "create":
        db.create_all()
        click.echo("Created...")
    elif action == "drop":
        db.drop_all()
        click.echo("Droped...")
    elif action == "upgrade":
        m = Migration(settings.SQL,
                      section=name,
                      alembic_file="alembic.ini")
        m.upgrade(
            to=to
        )

    elif action == "downgrade":
        m = Migration(settings.SQL,
                      section=name,
                      alembic_file="alembic.ini")

        m.downgrade(
            to=to
        )

    elif action == "revision":
        m = Migration(settings.SQL,
                      section=name,
                      alembic_file="alembic.ini")

        m.revision(
            rev_id=rev_id,
            message=message,
            head=head,
            branch_label=branch_label,
            depends_on=depends_on,
            autogenerate=True)
    else:
        console.print("[red bold]Wrong param...[/]")


@managercli.command(name="users")
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE, help="Fullpath to settings module")
@click.option("--is-superuser", "-S", default=False, help="Is a superuser")
@click.option("--scopes", "-C", default="user:r:w", help="Scopes for the user")
@click.argument("action", type=click.Choice(["create", "disable", "reset"]))
def userscli(action, is_superuser, scopes, settings_module):
    """Create a user"""

    settings = conf.load_conf(settings_module)
    if action == "create":
        name = Prompt.ask("Username")
        password = getpass("Password: ")
        repeat = getpass("Paswword (repeat): ")
        email = Prompt.ask("Email (optional)", default=None)
        if password != repeat:
            console.print("[bold red]Paswords doesn't match[/]")
            sys.exit(-1)

        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(create_user(
            settings, data={
                "username": name,
                "password": password,
                "email": email,
                "scopes": scopes,
                "is_superuser": is_superuser,
            }))
        if result:
            console.print(f"[bold magenta]Congrats!! user {name} created")
        else:
            console.print(f"[bold red]Something went wrong[/]")
            sys.exit(-1)

    # elif action == "reset":
    #     security = SecuritySettings()
    #     name = Prompt.ask("Username")
    #     _p = getpass("Password: ")
    #     with S() as session:
    #         changed = users_mg.change_pass(
    #             session, name, _p, salt=security.AUTH_SALT)
    #         session.commit()
    #     if changed:
    #         console.print("[bold magenta]Pasword changed[/]")
    #     else:
    #         console.print("[bold red]User may not exist [/]")
    # else:
    #     console.print("[red bold]Wrongs params[/]")



import os
from contextvars import ContextVar

from sanic import Request, Sanic
from sanic.log import logger
from sanic.response import HTTPResponse
from services import defaults
from services.base import PluginSpec
from services.db.sqlhelper import SQL, AsyncSQL
from services.types import Settings
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio.engine import AsyncEngine


class AsyncDBHelper:

    async def listener_db(self, app: Sanic):
        for k, v in app.config.DATABASES.items():
            _db = AsyncSQL(v,
                           set_session_factory=True)
            logger.info(f"Starting db {k}")
            await _db.init()
            app.ctx.databases[k] = _db

    def get_db(self, request, *, name="default") -> AsyncSQL:
        return request.app.ctx.databases.get(name)

    def get_session(self, request, *, name="default") -> AsyncSession:
        db = self.get_db(request, name=name)
        return db.session_factory()

    def get_engine(self, request, *, name="default") -> AsyncEngine:
        db = self.get_db(request, name=name)
        return db.engine


def init_db(app: Sanic, settings: Settings):
    app.config.DATABASES = settings.DATABASES
    db_helper = AsyncDBHelper()
    app.register_listener(db_helper.listener_db, "before_server_start")
    app.ext.dependency(db_helper)



import click
from rich.console import Console
from rich.prompt import Prompt
from services import conf, defaults, types, utils
from services.db import Migration
from services.db.sqlhelper import SQL


@click.group(name="db")
def dbcli():
    """
    DB related tasks
    """
    pass


@dbcli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--settings-module", "-s",
              default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
@click.argument("app_name")
def create(app_name, db, settings_module):
    """creates database schemas """
    settings = conf.load_conf(settings_module)

    db = SQL(settings.DATABASES[db])
    db.create_all(utils.get_meta_from_app(app_name))


@dbcli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--settings-module", "-s",
              default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
@click.argument("app_name")
def drop(app_name, db, settings_module):
    """drop database schemas """
    settings = conf.load_conf(settings_module)

    db = SQL(settings.DATABASES[db])
    db.drop_all(utils.get_meta_from_app(app_name))


@dbcli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini",
              help="alembic file")
@click.option("--to", "-t", default="head", help="Revision to upgrade or downgrade")
@click.option("--settings-module", "-s",
              default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
@click.argument("app_name")
def upgrade(app_name, db, settings_module, alembic, to):
    """upgrade a database based on the migrations created"""
    settings = conf.load_conf(settings_module)

    m = Migration(settings.DATABASES[db].sync_url,
                  section=app_name,
                  alembic_file=alembic)
    m.upgrade(
        to=to
    )


@dbcli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini",
              help="alembic file")
@click.option("--to", "-t", default="head", help="Revision to upgrade or downgrade")
@click.option("--settings-module", "-s",
              default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
@click.argument("app_name")
def downgrade(app_name, db, settings_module, alembic, to):
    """downgrade a database based on the migrations created"""
    settings = conf.load_conf(settings_module)

    m = Migration(settings.DATABASES[db].sync_url,
                  section=app_name,
                  alembic_file=alembic)
    m.downgrade(
        to=to
    )


@dbcli.command()
@click.option("--db", default="default", help="db name in the settings module")
@click.option("--alembic", "-a", default="alembic.ini")
@click.option("--message", "-m", default=None, help="Revision name")
@click.option("--rev-id", "-R", help="Revision id")
@click.option("--head", default=None, help="Specify head revision or <branchname>@head to base new revision.")
@click.option("--branch-label", "-b", default=None, help="Specify a branch label to apply to the new revision")
@click.option("--depends-on", "-d", help="list of 'depends on' identifiers")
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE, help="Fullpath to settings module")
@click.argument("app_name")
def revision(app_name, db, alembic,
             message,
             rev_id, head, branch_label, depends_on,
             settings_module):
    """ generates migrations files from models defined in apps """

    settings = conf.load_conf(settings_module)

    m = Migration(settings.DATABASES[db].sync_url,
                  section=app_name,
                  alembic_file=alembic)

    m.revision(
        rev_id=rev_id,
        message=message,
        head=head,
        branch_label=branch_label,
        depends_on=depends_on,
        autogenerate=True)



from sanic.exceptions import SanicException
from sanic.exceptions import Unauthorized as SanicUnauthorized


class SanicJWTException(SanicException):
    pass


class InvalidToken(SanicJWTException):
    pass


class WebAuthFailed(SanicJWTException):
    status_code = 401
    quiet = True

    def __init__(self, message="Authentication failed.", **kwargs):
        super().__init__(message, **kwargs)


class CommandExecutionException(Exception):
    def __init__(self, message):
        super().__init__(message)


class BadConfigurationException(Exception):
    def __init__(self, bad_param):
        super().__init__(f"{bad_param} is badly used")


class UserNotFound(Exception):

    def __init__(self, username):
        super().__init__(f"{username} not found")


class AuthValidationFailed(Exception):
    pass



from importlib import import_module
from typing import Callable

from sanic import Request, Sanic
from sanic.response import json

from services import defaults
from services.base import WebAppSpec
from services.db.plugin import init_db
from services.redis_conn import create_pool
from services.types import Settings
from services.utils import get_class, get_version

version = get_version("__version__.py")


async def status_handler(request):
    return json(dict(msg="We are ok", version=version))


# def init_db(app: Sanic, settings: Settings):
#    from services.db.web import sanic_init_db
#    sanic_init_db(app, settings)


# async def after_request(request: Request):
#     current_app = request.app
#     request.ctx.session = current_app.ctx.db.scoped_session()
#
#
# async def before_request(request: Request):
#     request.ctx.session.close()


def create_srv(
    settings: Settings,
    app_name=defaults.SANIC_APP,
    create_redis=create_pool,
    with_status_handler=True,
    auth_enabled=True,
    init_db_func: Callable = init_db,
) -> Sanic:
    """Factory pattern like flask"""

    app = Sanic(app_name)

    app.config.CORS_ORIGINS = settings.CORS_ORIGINS
    app.config.OAS_UI_DEFAULT = "swagger"
    app.config.OAS_UI_REDOC = False
    app.config.SWAGGER_UI_CONFIGURATION = {
        "apisSorter": "alpha",
        "operationsSorter": "alpha",
        "docExpansion": "none",
    }
    app.config.SETTINGS = settings
    if settings.REDIS:
        app.ctx.redis = create_redis(settings.REDIS)
    if settings.DATABASES:
        init_db(app, settings=settings)

    if auth_enabled:
        app.ext.openapi.add_security_scheme(
            "token",
            "http",
            scheme="bearer",
            bearer_format="JWT",
        )
        app.ext.openapi.secured("token")
        secure_app: WebAppSpec = get_class("services.security.web.WebApp")()
        secure_app.init(app, settings)

    app.ctx.databases = {}
    if settings.DATABASES:
        init_db_func(app, settings)

    for wapp in settings.APPS:
        w: WebAppSpec = get_class(wapp)()
        w.init(app, settings)

    if with_status_handler:
        app.add_route(status_handler, "/status")

    return app


def init_app(web_app_name: str, *, app: Sanic, settings: Settings):
    w: WebAppSpec = get_class(web_app_name)()
    w.init(app, settings)
    if w.bp_modules:
        w.init_blueprints(app)
    if w.middlewares:
        w.init_middlewares(app)



from sanic import Blueprint, Request
from sanic.response import json

{{ data.app_name }}_bp = Blueprint("{{ data.app_name }}", url_prefix="{{ data.app_name }}", version="v1")


@{{ data.app_name }}_bp.get("/")
# @openapi.response(200, {"application/json": JWTResponse})
async def default_handler(request: Request):
    return json(dict(msg="hello wolrd!"), 200)



import asyncio
import sys
from getpass import getpass
from typing import List

import click
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table
from services import conf, defaults, types
from services.db.sqlhelper import AsyncSQL
from services.security.authentication import auth_from_settings
from services.users import UserSpec
from services.users.managers import UserManager

console = Console()
loop = asyncio.get_event_loop()


def get_user_manager(settings: types.Settings) -> UserManager:
    db = AsyncSQL(settings.DATABASES[settings.USER_DB])
    auth = auth_from_settings(settings.SECURITY)
    # session = db.sessionmaker()
    manager = UserManager(auth, db=db, model_class=settings.USER_MODEL,
                          salt=settings.SECURITY.AUTH_SALT)
    return manager


async def _create_user(settings: types.Settings, data):
    manager = get_user_manager(settings)
    await manager.db.init()
    user = await manager.register(data)
    return user


async def _list_users(settings: types.Settings, active=True) -> List[UserSpec]:
    manager = get_user_manager(settings)
    await manager.db.init()
    users = await manager.list_users(active)
    return users


@click.group(name="users")
def userscli():
    """
    administration of users
    """
    pass


@userscli.command(name="create")
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
@click.option("--is-superuser", "-S", default=False, help="Is a superuser")
@click.option("--scopes", "-C", default="user:r:w", help="Scopes for the user")
def user_create(settings_module, is_superuser, scopes):
    settings = conf.load_conf(settings_module)
    name = Prompt.ask("Username")
    password = getpass("Password: ")
    repeat = getpass("Paswword (repeat): ")
    email = Prompt.ask("Email (optional)", default=None)
    if password != repeat:
        console.print("[bold red]Paswords doesn't match[/]")
        sys.exit(-1)
    result = loop.run_until_complete(_create_user(
        settings, data={
            "username": name,
            "password": password,
            "email": email,
            "scopes": scopes,
            "is_superuser": is_superuser,
        }))
    if result:
        console.print(f"[bold magenta]Congrats!! user {name} created")
    else:
        console.print(f"[bold red]Something went wrong[/]")
        sys.exit(-1)


@userscli.command(name="list")
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
def users_list(settings_module):
    settings = conf.load_conf(settings_module)

    result = loop.run_until_complete(_list_users(
        settings))
    table = Table(title="List of users")
    table.add_column("id")
    table.add_column("username")
    table.add_column("scopes")
    table.add_column("email")
    table.add_column("is_superuser")
    table.add_column("created at")

    for user in result:
        table.add_row(f"{user.id}",
                      user.username,
                      ",".join(user.scopes),
                      f"{user.email}",
                      f"{user.is_superuser}",
                      f"{user.create_at}")

    console.print(table)



import codecs
import importlib.util
import json
import os
import re
import subprocess
import unicodedata
from importlib import import_module
from pathlib import Path

from sqlalchemy.sql.schema import MetaData

from services import defaults
from services.errors import CommandExecutionException

_filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")
_windows_device_files = (
    "CON",
    "AUX",
    "COM1",
    "COM2",
    "COM3",
    "COM4",
    "LPT1",
    "LPT2",
    "LPT3",
    "PRN",
    "NUL",
)


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path="__version__.py"):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def get_query_param(request, key, default_val=None):
    val = request.args.get(key, default_val)
    return val


def mkdir_p(fp):
    """Make the fullpath
    similar to mkdir -p in unix systems.
    """
    Path(fp).mkdir(parents=True, exist_ok=True)


def open_json(path):
    with open(path, "r", encoding="utf-8") as f:
        data = f.read()
        dict_ = json.loads(data)
    return dict_


def execute_cmd(cmd) -> str:
    """Wrapper around subprocess"""
    with subprocess.Popen(
        cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as p:

        out, err = p.communicate()
        if err:
            raise CommandExecutionException(err.decode())

        return out.decode().strip()


def path_norm(fp):
    """Given a  filepath returns a normalized a path"""
    return str(Path(fp))


def get_parent_folder():
    """Get only the name of the parent folder
    commonly used to define the project name
    """
    root = Path(os.getcwd())
    return str(root).rsplit("/", maxsplit=1)[-1]


def get_package_dir(pkg):
    spec = importlib.util.find_spec(pkg)
    return spec.submodule_search_locations[0]


def get_from_module(fullpath):
    splited = fullpath.rsplit(".", maxsplit=1)
    mod_name = splited[0]
    obj_name = splited[1]
    mod = import_module(mod_name)
    obj = getattr(mod, obj_name)
    return obj


def secure_filename(filename: str) -> str:
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.
    On windows systems the function also makes sure that the file is not
    named after one of the special device files.
    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename('i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'
    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you abort or
    generate a random filename if the function returned an empty one.
    .. versionadded:: 0.5
    :param filename: the filename to secure
    """
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.encode("ascii", "ignore").decode("ascii")

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip(
        "._"
    )

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if (
        os.name == "nt"
        and filename
        and filename.split(".")[0].upper() in _windows_device_files
    ):
        filename = f"_{filename}"

    return filename


def normalize_name(name: str) -> str:
    """used mostly for projects"""
    evaluate = name.lower()
    evaluate = evaluate.replace(" ", "_")
    evaluate = secure_filename(name)
    return evaluate


def get_class(fullclass_path):
    """get a class or object from a module. The fullclass_path should be passed as:
    package.my_module.MyClass
    """
    module, class_ = fullclass_path.rsplit(".", maxsplit=1)
    mod = import_module(module)
    cls = getattr(mod, class_)
    return cls


def init_blueprints(app, blueprints_allowed, package_dir="services.web"):
    """
    It will import bluprints from modules that ends with "_bp" and belongs
    to the package declared in `package_dir`
    """
    blueprints = set()
    mod = app.__module__
    for mod_name in blueprints_allowed:
        modules = import_module(f"{package_dir}.{mod_name}", mod)
        for el in dir(modules):
            if el.endswith("_bp"):
                bp = getattr(modules, el)
                blueprints.add(bp)

    for bp in blueprints:
        print("Adding blueprint: ", bp.name)
        app.blueprint(bp)


def get_meta_from_app(app_name) -> MetaData:
    Base = get_class(f"{app_name}.db.Base")
    return Base.metadata



import os

import click
from rich.console import Console

from services.commands.common import create_app
from services.commands.db import dbcli
from services.commands.web import webcli
from services.commands.users import userscli


def init_cli():

    console = Console()

    @click.group()
    def cli():
        """
        Command line tool
        """

    @click.command()
    def version():
        """Actual version"""
        from services.utils import get_version

        ver = get_version("__version__.py")
        console.print(f"[bold magenta]{ver}[/bold magenta]")

    cli.add_command(version)
    cli.add_command(webcli)
    cli.add_command(userscli)
    cli.add_command(dbcli)
    cli.add_command(create_app)
    return cli


cli = init_cli()

if __name__ == "__main__":

    cli()



from functools import wraps
from inspect import isawaitable
from typing import List, Optional

from sanic import Request, Sanic, json
from services.base import WebAppSpec
from services.security.authentication import auth_from_settings
from services.security.redis_tokens import RedisTokenStore
from services.types import Settings

from .base import AuthSpec
from .errors import (AuthValidationFailed, MissingAuthorizationHeader,
                     WebAuthFailed)


def get_req_auth(request: Request) -> AuthSpec:
    """a shortcut to get the Auth object from a web context"""

    return request.app.ctx.auth


def get_app_auth(app: Sanic) -> AuthSpec:
    """a shortcut to get the Auth object from a web context"""

    return app.ctx.auth


def protected(scopes: Optional[List[str]] = None, require_all=True):
    """verify a token from a request.
    Optionally if a list of scopes is given then will check that scopes
    with the scopes provided by the token.

    :param scopes: a list of scopes
    :param required_all: if true it will check that the all the names provided
    match with the required.

    """

    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            auth = get_req_auth(request)
            token = request.token
            if not token:
                raise MissingAuthorizationHeader()
            try:
                decoded = auth.validate(token, scopes, require_all)
                request.ctx.token_data = decoded
                response = f(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response
            except AuthValidationFailed:
                raise WebAuthFailed()

            return response

        return decorated_function

    return decorator


def auth_error_handler(request, exception):
    return json({"msg": "Authentication failed"}, status=401)


class WebApp(WebAppSpec):
    name = "security"
    package_dir = "services.security"

    def init(self, app: Sanic, settings: Settings):
        store = None
        if settings.SECURITY.AUTH_ALLOW_REFRESH:
            try:
                store = RedisTokenStore(app.ctx.web_redis)
            except AttributeError:
                print("NO redis configuration")

        app.ctx.auth = auth_from_settings(settings.SECURITY, store)
        app.config.AUTH_SALT = settings.SECURITY.AUTH_SALT
        app.config.AUTH_ALLOW_REFRESH = settings.SECURITY.AUTH_ALLOW_REFRESH

        app.error_handler.add(WebAuthFailed, auth_error_handler)
        app.error_handler.add(MissingAuthorizationHeader, auth_error_handler)



from sanic import Sanic
from services.base import WebAppSpec
from services.security import get_app_auth
from services.types import Settings
{% if data.init %}
from services.users import UserManager

class WebApp(WebAppSpec):
    name = "{{ data.app_name }}"
    bp_modules = ["views_bp", "users_bp"]
    package_dir = "{{ data.app_name }}"

    def hook_users(self, app: Sanic):
        """
        delete if you don't want to use the pre-built user system
        """
        settings = self.get_app_settings(app)
        db = self.get_db(app, settings.USER_DB)
        auth = get_app_auth(app)

        app.ctx.users_mg = UserManager(
            auth,
            db=db,
            model_class=settings.USER_MODEL,
            salt=settings.SECURITY.AUTH_SALT,
        )

    def init(self, app: Sanic, settings: Settings):
        """ complete with your own logic """
        app.register_listener(self.hook_users, "before_server_start")

        self.init_blueprints(app)
{% else %}

class WebApp(WebAppSpec):
    name = "{{ data.app_name }}"
    bp_modules = ["views"]
    package_dir = "{{ data.app_name }}"

    def init(self, app: Sanic, settings: Settings):
        """ complete with your own logic """
        pass

{% endif %}



import sys

from services import defaults, types

# WARNING:
# We do our best effort to keep sensible information private
# but in the scenario of an intrusion into the network or machines
# where servers are hosted, then it represents a risk for the information
# stored in those machines.

# General Folders for the server
SECURITY = types.SecuritySettings(
    JWT_SECRET="Changeme",
    JWT_ALG="HS256",
    AUTH_SALT="Changeme",
)

USER_MODEL = "{{ data.app_name }}.models.UserModel"

DATABASES = {
    "default": types.Database(
        name="default",
        async_url="sqlite+aiosqlite:///db.sqlite",
        sync_url="sqlite:///db.sqlite",
        description="Default database",
    )
}

APPS = [
    # "services.users.web.WebApp",
    "services.security.web.WebApp",
    "{{ data.app_name }}.web.WebApp",
]



from functools import wraps
from inspect import isawaitable
from typing import List, Optional

from sanic import Request, Sanic, json
from services.base import WebAppSpec
from services.security import get_app_auth
from services.security.authentication import auth_from_settings
from services.types import Settings

from .managers import UserManager


def get_users_mg(request: Request) -> UserManager:
    """a shortcut to get the Auth object from a web context"""
    current_app: Sanic = Sanic.get_app(request.app.name)

    return current_app.ctx.users_mg



from sqlalchemy import inspect
from sqlalchemy.schema import (DropConstraint, DropTable, ForeignKeyConstraint,
                               MetaData, Table)


def drop_everything(engine):
    """(On a live db) drops all foreign key constraints before dropping all tables.
    Workaround for SQLAlchemy not doing DROP ## CASCADE for drop_all()
    (https://github.com/pallets/flask-sqlalchemy/issues/722)
    """

    con = engine.connect()
    trans = con.begin()
    inspector = inspect(engine)

    # We need to re-create a minimal metadata with only the required things to
    # successfully emit drop constraints and tables commands for postgres (based
    # on the actual schema of the running instance)
    meta = MetaData()
    tables = []
    all_fkeys = []

    for table_name in inspector.get_table_names():
        fkeys = []

        for fkey in inspector.get_foreign_keys(table_name):
            if not fkey["name"]:
                continue

            fkeys.append(ForeignKeyConstraint((), (), name=fkey["name"]))

        tables.append(Table(table_name, meta, *fkeys))
        all_fkeys.extend(fkeys)

    for fkey in all_fkeys:
        con.execute(DropConstraint(fkey))

    for table in tables:
        con.execute(DropTable(table))

    trans.commit()



import os
from contextvars import ContextVar

from services import defaults
from services.db.nosync import AsyncSQL
from services.types import Settings
from sanic import Request, Sanic
from sanic.response import HTTPResponse

_base_model_session_ctx = ContextVar("session")


def create_db_instance(url) -> AsyncSQL:
    return AsyncSQL(url)


async def inject_session_req(request: Request):
    current_app = request.app
    request.ctx.session = current_app.ctx.db.sessionmaker()
    request.ctx.session_ctx_token = _base_model_session_ctx.set(
        request.ctx.session)


async def close_session_res(request: Request, response: HTTPResponse):
    if hasattr(request.ctx, "session_ctx_token"):
        _base_model_session_ctx.reset(request.ctx.session_ctx_token)
        await request.ctx.session.close()


async def listener_db(app: Sanic):
    _db = create_db_instance(app.config.ASQL)
    app.ctx.db = _db
    await app.ctx.db.init()


def sanic_init_db(app: Sanic, settings: Settings):
    app.config.ASQL = settings.ASQL
    app.register_listener(listener_db, "before_server_start")
    app.register_middleware(inject_session_req, "request")
    app.register_middleware(close_session_res, "response")



# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict

from services.db.common import Base
from sqlalchemy import (BigInteger, Boolean, Column, DateTime, Float,
                        ForeignKey, Integer, String, UniqueConstraint)
# from sqlalchemy.dialects.postgresql import BYTEA
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property
from sqlalchemy.orm import declarative_mixin, declared_attr, relationship
from sqlalchemy.schema import Table
from sqlalchemy.sql import functions
from sqlalchemy.types import BINARY


@declarative_mixin
class UserMixin:

    id = Column(BigInteger, primary_key=True)
    username = Column(String(), index=True, unique=True, nullable=False)
    password = Column("password", BINARY, nullable=True)
    email = Column(String(), index=True, nullable=True, unique=True)
    is_superuser = Column(Boolean, default=False, nullable=False)
    is_active = Column(Boolean, default=True, nullable=False)
    scopes = Column(String(), default="user", nullable=False)

    created_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )
    updated_at = Column(
        DateTime(),
        server_default=functions.now(),
        nullable=False,
    )

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError()




from pathlib import Path

import click
from rich.console import Console
from rich.panel import Panel
from services import init_script

console = Console()


@click.command(name="create-app")
@click.argument("appname")
def create_app(appname):
    """Create the structure of an app"""
    root = Path.cwd()
    init_script.create_app(root, appname)
    console.print(f"[green bold]App {appname} created.[/]")



from abc import ABC, abstractmethod
from typing import Callable, List, Optional, Tuple

from pydantic import BaseModel
from sanic import HTTPResponse, Request, Sanic

from services.db.sqlhelper import AsyncSQL
from services.security.base import AuthSpec
from services.types import Settings
from services.utils import init_blueprints


class WebAppSpec(ABC):
    """ Spec whichs define a WebApp.
    A WebApp can have endpoints, or not.
    """
    name: str
    bp_modules: List[str]
    package_dir: str
    # middlewares: List[Tuple[str, Callable]]
    # dependencies: List[Callable]

    @abstractmethod
    def init(self, app: Sanic, settings: Settings):
        pass

    def init_blueprints(self, app: Sanic):
        init_blueprints(app, self.bp_modules, package_dir=self.package_dir)

    def get_db(self, app: Sanic, name="default") -> AsyncSQL:
        return app.ctx.databases[name]

    def get_app_settings(self, app: Sanic) -> Settings:
        return app.config.SETTINGS

    def get_request_settings(self, request: Request) -> Settings:
        return request.app.config.SETTINGS


class PluginSpec(ABC):
    name: str
    request_ctx_name: Optional[str] = None
    request_hook_enabled: bool = False
    response_hook_enabled: bool = False

    @abstractmethod
    def init(self, app: Sanic, settings: Settings):
        pass

    def get_from_request(self, request):
        obj = getattr(request.app.ctx, self.name)
        return obj

    @abstractmethod
    async def request_hook(self, request: Request):
        raise NotImplementedError("Response hook not implemented")

    @abstractmethod
    async def response_hook(self, request: Request, response: HTTPResponse):
        raise NotImplementedError("Response hook not implemented")



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field
from services.db.sqlhelper import AsyncSQL
from services.security import AuthSpec, PasswordScript
from services.types import JWTResponse
from services.utils import get_class


class UserSpec(BaseModel):
    id: int
    username: str
    scopes: List[str] = ["user"]
    email: Optional[str] = None
    is_superuser: bool = False
    is_active: bool = True
    create_at: datetime = datetime.utcnow()
    updated_at: datetime = datetime.utcnow()


class UserManagerSpec(ABC):
    auth: AuthSpec

    @abstractmethod
    async def register(self, data: Dict[str, Any]) -> Union[UserSpec, None]:
        pass

    @abstractmethod
    async def update(self, username: str, data: Dict[str, Any]):
        pass

    @abstractmethod
    async def list_users(self, filter_active=True) -> List[UserSpec]:
        pass

    @abstractmethod
    async def get(self, username: str) -> UserSpec:
        pass

    @abstractmethod
    async def authenticate(self, username: str, *,
                           password: str) -> UserSpec:
        pass

    @abstractmethod
    def verify_password(self,
                        original_password: bytes, *,
                        to_verify: str,
                        salt: Union[bytes, str]
                        ) -> bool:
        pass

    @abstractmethod
    def encrypt_password(self, password, salt: Union[bytes, str]) -> bytes:
        pass

    @abstractmethod
    async def delete(self, username: str):
        pass

    @abstractmethod
    async def change_pass(self, username: str, new_password: str):
        pass

    @abstractmethod
    async def generate_token(self, user: UserSpec, exp=None) -> JWTResponse:
        pass

    @abstractmethod
    async def refresh_token(self, old_token: JWTResponse) -> JWTResponse:
        pass



from sanic import Sanic
from services.base import MigrationSpec, WebAppSpec
from services.types import Settings


class WebApp(WebAppSpec):
    name = "{{ data.app_name }}"
    bp_modules = ["views"]
    package_dir = "{{ data.app_name }}"

    def init(self, app: Sanic, settings: Settings):
        """ complete with your own logic """
        pass



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from services.db.sqlhelper import AsyncSQL
from services.errors import UserNotFound, AuthValidationFailed
from services.security import AuthSpec, PasswordScript
from services.types import JWTConfig, JWTResponse
from services.utils import get_class
from sqlalchemy import delete, select
from sqlalchemy.exc import IntegrityError

from .base import UserManagerSpec, UserSpec
from .models import UserMixin


class UserManager(UserManagerSpec):

    def __init__(self, auth: AuthSpec,
                 *,
                 db: AsyncSQL,
                 model_class: str,
                 salt: Optional[str] = None):
        self._model: UserMixin = get_class(model_class)
        self._salt = salt
        self.db = db
        self.auth = auth

    def verify_password(self,
                        original_password: bytes, *,
                        to_verify: str,
                        salt: Union[bytes, str]
                        ) -> bool:
        pm = PasswordScript(salt)
        verified = pm.verify(to_verify, original_password)
        if verified:
            return True
        return False

    def encrypt_password(self, password, salt: Union[bytes, str]) -> bytes:

        pm = PasswordScript(salt)
        encrypted = pm.encrypt(password)
        return encrypted

    async def authenticate(self, username: str, *,
                           password: str) -> UserSpec:

        async with self.db.session_factory() as session:
            user = await self._get(session, username)
            if user is None:
                raise AuthValidationFailed()

            is_valid = self.verify_password(
                user.password, to_verify=password, salt=self._salt
            )
            if not is_valid or not user.is_active:
                raise AuthValidationFailed()

            return self._to_spec(user)

    def _to_spec(self, user: UserMixin) -> UserSpec:
        return UserSpec(
            id=user.id,
            username=user.username,
            email=user.email,
            is_superuser=user.is_superuser,
            is_active=user.is_active,
            scopes=user.scopes.split(","),
            created_at=user.created_at,
            updated_at=user.updated_at,
        )

    def _user_or_raise(self, username,
                       user_object: Union[UserMixin, None]) -> UserSpec:
        if not user_object:
            raise UserNotFound(username)
        return self._to_spec(user_object)

    async def list_users(self, filter_active=True) -> List[UserSpec]:
        stmt = select(self._model)
        if filter_active:
            stmt.where(self._model.is_active == True)
        async with self.db.session_factory() as session:
            result = await session.execute(stmt)
            return [self._to_spec(r[0]) for r in result]

    async def get(self, username: str) -> UserSpec:

        stmt = select(self._model)\
            .where(self._model.username == username)\
            .limit(1)

        async with self.db.session_factory() as session:
            rsp = await session.execute(stmt)

            _user = rsp.scalar_one_or_none()
        return self._user_or_raise(username, _user)

    async def _get(self, session, username: str) -> UserMixin:
        stmt = select(self._model)\
            .where(self._model.username == username)\
            .limit(1)

        rsp = await session.execute(stmt)

        return rsp.scalar_one_or_none()

    async def register(self, data: Dict[str, Any]) -> Union[UserSpec, None]:

        data["password"] = self.encrypt_password(data["password"], self._salt)

        um = self._model(**data)

        async with self.db.session_factory() as session:
            session.add(um)
            try:
                await session.commit()
                return self._to_spec(um)
            except IntegrityError:
                await session.rollback()
            return None

    async def update(self, username: str, data: Dict[str, Any]):
        async with self.db.session_factory() as session:
            u = await self._get(session, username)
            if not u:
                raise UserNotFound(username)

            for k, v in data.items():
                setattr(u, k, v)
            u.updated_at = datetime.utcnow()
            session.add(u)

    async def delete(self, username: str):

        async with self.db.session_factory() as session:
            stmt = delete(self._model).where(self._model.username == username)
            await session.execute(stmt)

    async def change_pass(self, session, username: str, new_password: str):
        async with self.db.session_factory() as session:
            u = await self._get(session, username)
            if not u:
                raise UserNotFound(username)
            pass_ = self.encrypt_password(new_password, salt=self._salt)
            u.password = pass_
            u.updated_at = datetime.utcnow()
            session.add(u)

    async def generate_token(self, user: UserSpec, exp=None) -> JWTResponse:

        payload = {
            "usr": user.username, "scopes": user.scopes
        }
        encoded = self.auth.encode(payload)
        rftkn = None
        if self.auth.store:
            rftkn = await self.auth.store_refresh_token(user.username)
        return JWTResponse(
            access_token=encoded,
            refresh_token=rftkn
        )

    async def refresh_token(self, old_token: JWTResponse) -> JWTResponse:
        new_jwt = await self.auth.refresh_token(
            old_token.access_token, old_token.refresh_token)
        return new_jwt



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['services',
 'services.commands',
 'services.db',
 'services.security',
 'services.templates',
 'services.templates.alembic',
 'services.templates.app',
 'services.users']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyJWT>=2.4.0,<3.0.0',
 'SQLAlchemy>=1.4.39,<2.0.0',
 'alembic>=1.8.1,<2.0.0',
 'cryptography>=37.0.4,<38.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'sanic[ext]>=22.6.0,<23.0.0']

entry_points = \
{'console_scripts': ['create-srv-project = '
                     'services.cli_project:create_service_project',
                     'srv = services.cli:cli']}

setup_kwargs = {
    'name': 'ai-services',
    'version': '0.3.0',
    'description': 'A simple web framework based on Sanic',
    'long_description': None,
    'author': 'nuxion',
    'author_email': 'nuxion@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nuxion/services',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from typing import Any, Dict

{% if data.init %}
from services.users.models import UserMixin
{% endif %}
from sqlalchemy import BigInteger, Column, String

from .db import Base


{% if data.init %}
class UserModel(UserMixin, Base):
    __tablename__ = "{{ data.app_name }}_user"
    __mapper_args__ = {"eager_defaults": True}

    fullname = Column(String(), nullable=True)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "username": self.username,
            "password": self.password,
            "fullname": self.fullname,
            "email": self.email,
            "is_superuser": self.is_superuser,
            "is_active": self.is_active,
            "scopes": self.scopes,
            "created_at": self.created_at,
            "updated_at": self.updated_at
        }
{% endif %}

class ExampleModel(Base):

    __tablename__ = "{{ data.app_name }}_example"
    __mapper_args__ = {"eager_defaults": True}

    id = Column(BigInteger, primary_key=True)
    name = Column(String())



from pathlib import Path

import click
from rich.console import Console
from rich.panel import Panel

from services import init_script

console = Console()


@click.command()
@click.argument("base_path")
def create_service_project(base_path):
    """ Start a new project """
    root = Path(base_path).resolve()

    p = Panel.fit(
        "[bold magenta]:smile_cat: Hello and welcome to "
        " AI services [/bold magenta]",
        border_style="red",
    )
    console.print(p)

    default_app = init_script.ask_webapp_name()
    init_script.create_default(root, default_app)



import os

import click
from services import defaults


@click.command(name="web")
@click.option("--host", "-H", default="127.0.0.1", help="Listening Host")
@click.option("--port", "-p", default="8000", help="Listening Port")
@click.option("--workers", "-w", default=1, help="How many workers start?")
@click.option(
    "--reload", "-r", default=False, is_flag=True, help="Enable Auto reload"
)
@click.option(
    "--access-log", "-L", default=False, is_flag=True, help="Enable access_log"
)
@click.option("--debug", "-D", default=False, is_flag=True, help="Enable Auto reload")
@click.option("--settings-module", "-s", default=defaults.SETTINGS_MODULE,
              help="Fullpath to settings module")
def webcli(host, port, workers, reload, access_log, debug, settings_module):
    """Run Web Server"""
    # pylint: disable=import-outside-toplevel
    from services.conf import load_conf
    from services.server import create_srv

    # from playpy.web import sanic_init_indexer
    # from playpy.types import Settings as PlaySettings
    pwd = os.getcwd()


    settings = load_conf(settings_module)

    host = host or settings.HOST
    port = port or settings.PORT
    w = int(workers)

    print(f"BASE_PATH: {settings.BASE_PATH}")
    print(f"SETTINGS_MODULE: {settings.SETTINGS_MODULE}")
    print(f"Debug mode: {debug}")
    print(f"Access log: {access_log}")
    print(f"Autoreload: {reload}")
    print(f"Workers: {w}")
    print(f"Listening: {host}:{port}")

    srv = create_srv(settings)
    srv.run(
        host=host,
        port=int(port),
        access_log=access_log,
        workers=w,
        debug=debug,
        auto_reload=reload,
    )



import sys

from services import defaults, types
# WARNING:
# We do our best effort to keep sensible information private
# but in the scenario of an intrusion into the network or machines
# where servers are hosted, then it represents a risk for the information
# stored in those machines.


# General Folders for the server
SECURITY = types.SecuritySettings(
    JWT_SECRET="Changeme",
    JWT_ALG="HS256",
    AUTH_SALT="Changeme",
)
APPS = [
    "services.users.web.WebApp",
]

# Logs
# LOGLEVEL = "INFO"
# LOGCONFIG = dict(  # no cov
#     version=1,
#     # disable_existing_loggers=True,
#     loggers={
#         defaults.LOGGER: {
#             "level": LOGLEVEL,
#             "handlers": ["console"],
#         },
#     },
#     handlers={
#         "console": {
#             "class": "logging.StreamHandler",
#             "formatter": "generic",
#             "stream": sys.stdout,
#         },
#         "error_console": {
#             "class": "logging.StreamHandler",
#             "formatter": "generic",
#             "stream": sys.stderr,
#         },
#     },
#     formatters={
#         "generic": {
#             # "format": "%(asctime)s [%(process)d] [%(levelname)s] %(message)s",
#             "format": "[%(levelname)s] %(message)s",
#             "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
#             "class": "logging.Formatter",
#         },
#     },
# )



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, BaseSettings, RedisDsn

from services import defaults


class Database(BaseModel):
    """
    Database configuration object to be used with
    :class:`services.db.sqlhelper.SQL` and
    :class:`services.db.sqlhelper.AsyncSQL`

    Examples:

    .. highlight:: python
    .. code-block:: python

        from services.types import Database
        dbconf = Database(uri="sqlite:////tmp/test.db")


    :param uri: full dsl url sqlite:////tmp/test.db
    :type uri: str
    :param name: optional value to identify a database
    :type name: str
    :param pool_size: size of the pool
    :type pool_size: int
    :param max_overflow: TODO
    :type max_overflow: int
    :param debug: maps with echo param in sqlalachemy
    :param description: optional value for metadata info about the database

    """
    async_url: Optional[str] = None
    sync_url: Optional[str] = None
    name: str = "default"
    pool_size: int = 20
    max_overflow: int = 0
    debug: bool = False  # debug
    # metadata
    description: Optional[str] = None


class MigrationType(BaseModel):
    models: List[str]
    package_dir: str
    version_table: str


class SecuritySettings(BaseSettings):
    JWT_PUBLIC: Optional[str] = None
    JWT_PRIVATE: Optional[str] = None
    JWT_SECRET: Optional[str] = None
    JWT_ALG: str = "ES512"
    JWT_EXP: int = 30  # 30 minutes
    JWT_CLAIMS_REQUIRED: List[str] = ["exp"]
    JWT_SECRET: Optional[str] = None
    JWT_ISS: Optional[str] = None
    JWT_AUD: Optional[str] = None
    REFRESH_TOKEN_TTL: int = 3600 * 168  # 7 days
    AUTH_SALT: str = "changeit"
    AUTH_ALLOW_REFRESH: bool = True
    AUTH_CLASS = "services.security.authentication.Auth"

    class Config:
        env_prefix = "SRV_"


class Settings(BaseSettings):
    BASE_PATH: Union[str, Path]
    HOST: str = "localhost"
    PORT: str = "8000"
    DATABASES: Dict[str, Database] = {}
    REDIS: Optional[RedisDsn] = None
    REDIS_POOL_SIZE: int = 10
    TEMPLATES_DIR: Optional[str] = None
    TEMPLATES_PACKAGE_NAME: Optional[str] = None
    DEV_MODE: bool = False
    SECURITY: Optional[SecuritySettings] = None
    CUSTOM_COMMANDS: List[str] = []
    USER_ENDPOINTS: bool = True
    USER_MODEL: Optional[str] = None
    USER_DB: str = "default"
    USER_MANAGER_CLASS = "services.user.managers.UserManager"

    APPS: List[str] = []

    # logs
    LOGLEVEL: str = "INFO"
    LOGCONFIG: Dict[str, Any] = {}
    DEBUG: bool = False

    # MIGRATIONS: Dict[str, Migration] = {}

    CORS_ORIGINS: Union[List, str] = "*"
    CORS_ALLOW_HEADERS: Union[List, str] = "*"
    SANIC_APP_NAME = defaults.SANIC_APP
    SETTINGS_MODULE: Optional[str] = None

    class Config:
        env_prefix = "SRV_"


class KeyPairs(BaseModel):
    public: str
    private: str


class JWTConfig(BaseModel):
    alg: str
    exp_min: int = 30
    keys: Optional[KeyPairs] = None
    secret: Optional[str] = None
    issuer: Optional[str] = None
    audience: Optional[str] = None
    requires_claims: List[str] = ["exp"]
    ttl_refresh_token: int = 3600 * 168  # 7 days

    class Config:
        extra = "forbid"


class UserLogin(BaseModel):
    username: str
    password: str

    class Config:
        extra = "forbid"


class JWTResponse(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None


class TokenCreds(BaseModel):
    access_token: str
    refresh_token: Optional[str] = None

    class Config:
        extra = "forbid"



from .common import Base
from .migration import Migration
from .pages import (NextPage, Pagination, get_total, get_total_async,
                    get_total_table, get_total_table_async)
from .sqlhelper import SQL, AsyncSQL
